package fit;

import fit.asympdist.AsympDist;
import fit.growth.Growth;
import fit.utils.DoublePair;
import java.util.Arrays;
import java.util.stream.IntStream;

/**
 * Class for fitting a model.
 */
public class Fit {

    /**
     * If an infeasible solution was produced, this returns positive infinity.
     */
    public static final double INFEASIBLE = Double.POSITIVE_INFINITY;

    /**
     * Data for estimation. Array of {{age, value},...}
     */
    double[][] TX;

    /**
     * Lower bounds of parameters to be estimated.
     */
    public final double[] Min;

    /**
     * Upper bounds of parameters to be estimated.
     */
    public final double[] Max;

    /**
     * Number of parameters to be estimated, except for distribution parameters of asymptotes.
     */
    public final int numCurveParams;

    /**
     * Number of parameters to be estimated, including distribution parameters of asymptotes.
     */
    public final int numTotalParams;

    /**
     * Model of the distribution of asymptotes.
     */
    public final AsympDist asympDist;

    /**
     * Model of the growth curve set.
     */
    public final Growth growth;

    /**
     * Constructor.
     *
     * @param _growth The model of growth curve set.
     * @param _asympDist The model of the distribution of asymptotes.
     * @param _Vmin Lower bounds of parameters, except for distribution parameters of asmptotes.
     * @param _Vmax Uppwer bounds of parameters, except for distribution parameters of asmptotes.
     */
    public Fit(Growth _growth, AsympDist _asympDist, double[] _Vmin, double[] _Vmax) {
        if (_Vmin.length != _Vmax.length || _Vmax.length != _growth.paramLen()) {
            throw new IllegalArgumentException("Illegal length.");
        }
        growth = _growth;
        asympDist = _asympDist;
        Min = _Vmin;
        Max = _Vmax;
        numCurveParams = _growth.paramLen();
        numTotalParams = _growth.paramLen() + _asympDist.paramLen();
    }

    /**
     * The array is deep-copied.
     *
     * @param _TX Array to set.
     */
    public final void setData(double[][] _TX) {
        TX = new double[_TX.length][2];
        IntStream.range(0, TX.length).forEach(i -> System.arraycopy(_TX[i], 0, TX[i], 0, TX[i].length));
    }

    /**
     * Returns AIC.
     *
     * @param _lnLH Log LH value.
     * @return AIC.
     */
    public final double getAIC(double _lnLH) {
        return 2 * (-_lnLH + numTotalParams);
    }

    /**
     * Returns asymptotes for all data in TX with growth model parameters.
     *
     * @param _growthParams Growth model parameters.
     * @return Asymptotes.
     */
    public final double[] getAsymptotes(double[] _growthParams) {
        return IntStream.range(0, TX.length).parallel().mapToDouble(i -> growth.getAsymptote(TX[i][0], TX[i][1], _growthParams)).toArray();
    }

    /**
     * Objective function to be minimized.
     *
     * @param _Params Parameters of growth and survive.
     * @return Negative log likelihood.
     */
    public double f(double... _Params) {
        assert (_Params.length == numCurveParams);
        double[] Asymptotes = getAsymptotes(_Params);
        if (Arrays.stream(Asymptotes).anyMatch(d -> Double.isInfinite(d) || Double.isNaN(d))) {
            return INFEASIBLE;
        }
        DoublePair AsymDist = asympDist.MLE(Asymptotes);
        double ll = 0;
        for (int i = 0; i < TX.length; i++) {
            ll += asympDist.logPDF(Asymptotes[i], AsymDist.params)
                    + growth.log_dAdX(TX[i][0], Asymptotes[i], _Params);
        }
        return Double.isNaN(ll) ? INFEASIBLE : -ll;//Minimized, so negative lm
    }

    /**
     * Returns a String that identify the model.
     *
     * @return String that indicate the site index model.
     */
    public String getType() {
        return "Growth=" + growth.getClass().getSimpleName() + ", Asymptote=" + asympDist.getClass().getSimpleName();
    }

    /**
     * Returns the growth curves with the percentile value of asymptote.
     *
     * @param _T Ages to predict
     * @param _AsympDistParams Distribution parameters of asymototes
     * @param _growthParams Growth parameters.
     * @param _percentile01 Percentile value in the range of (0, 1) to calculate.
     * @return Array of sizes at each t.
     */
    public double[] getEstimates(int[] _T, double[] _AsympDistParams, double[] _growthParams, double _percentile01) {
        double asymptote = asympDist.InvCDF(_percentile01, _AsympDistParams);
        return Arrays.stream(_T).mapToDouble(t-> growth.calcXAtT(t, asymptote, _growthParams)).toArray();
    }
}
