package fit.asympdist;

import fit.utils.DoublePair;
import java.util.Arrays;
import org.apache.commons.math3.analysis.solvers.BrentSolver;
import org.apache.commons.math3.distribution.GammaDistribution;
import org.apache.commons.math3.exception.NoBracketingException;
import org.apache.commons.math3.special.Gamma;
import static org.apache.commons.math3.special.Gamma.digamma;
import static org.apache.commons.math3.special.Gamma.trigamma;

/**
 * Generalized gamma distribution.
 */
public class AsympDist_GGD implements AsympDist {

    /**
     * Maximum number of iteration for the estimation of GGD parameters.
     */
    public final static int MAX_ITER = 10000;

    @Override
    public int paramLen() {
        return 3;
    }

    /**
     * Inversed CDF. _Params[0]: First shape parameter (compatible with the shape parameter of normal gamma distribution). _Params[1]: Second shape parameter
     * (Th exponential value of random variable). _Params[2]: Scale parameter.
     *
     * @param _p Cumulative probability value.
     * @param _Params Prameters.
     * @return The value inverse CDF.
     */
    @Override
    public double InvCDF(double _p, double... _Params) {
        return Math.pow(
                new GammaDistribution(_Params[0], Math.pow(_Params[2], _Params[1])).inverseCumulativeProbability(_p),
                1 / _Params[1]);
    }

    /**
     * Returns MLE-estimated GGD parameters and maximum log likelihood. Excecutes MLE of gamma distribution for s-powered variables, changing the value of s.
     *
     * @param _V Random variables.
     * @return logML, {firstShape, secondShape, scale}.
     */
    @Override
    public DoublePair MLE(double[] _V) {
        BrentSolver bs = new BrentSolver();
        double[] Best = new double[3];
        double[] b = {Double.NEGATIVE_INFINITY};
        try {
            bs.solve(MAX_ITER, s -> {
                double[] Vs = Arrays.stream(_V).map(v -> Math.pow(v, s)).toArray();
                DoublePair Res = gammaMLE(Vs);
                if (Double.isInfinite(Res.llh)) {
                    return Double.NEGATIVE_INFINITY;
                }
                double l = Res.llh;
                double[] R = Res.params;
                double meanOfLogRaw = Arrays.stream(_V).map(v -> Math.log(v)).average().getAsDouble();
                double meanOfLogTrans = Arrays.stream(Vs).map(v -> Math.log(v)).average().getAsDouble();
                double ml = _V.length * (Math.log(s) + (R[0] * s - 1) * meanOfLogRaw);
                double logL_old = _V.length * ((R[0] - 1) * meanOfLogTrans);
                l += ml - logL_old;
                double scale = Math.pow(R[1], 1 / s);
                if (b[0] < l) {
                    b[0] = l;
                    Best[0] = R[0];
                    Best[1] = s;
                    Best[2] = scale;
                }
                return 1 / s + R[0] * meanOfLogRaw
                        - Arrays.stream(_V).map(v -> Math.pow(v / scale, s - 1)).average().getAsDouble()
                        - R[0] * Math.log(scale);
            }, 0.1, 10, 1);//Optimum second shape is searched in range [0.1, 10] with initial value 1.
        } catch (StackOverflowError | NoBracketingException e) {
            Arrays.fill(Best, Double.NaN);//Returns NaN if the solver failed.
        }
        return new DoublePair(b[0], Best);
    }

    /**
     * Returns logarithmic PDF.
     *
     * @param _x Random variable.
     * @param _Params {shape1, shape2, scale}
     * @return logPDF
     */
    @Override
    public double logPDF(double _x, double... _Params) {
        return Math.log(_Params[1] / _Params[2]) - Gamma.logGamma(_Params[0])
                + (_Params[0] * _Params[1] - 1.) * Math.log(_x / _Params[2])
                - Math.pow(_x / _Params[2], _Params[1]);
    }

    /**
     * MLE of gamma distribution, using Minka (2002)'s method. See https://tminka.github.io/papers/minka-gamma.pdf
     *
     * @param _V Random variables.
     * @return {logML, shape, scale}.
     */
    private DoublePair gammaMLE(double[] _V) {
        double logOfMean, meanOfLog = 0, mean = 0;
        for (double v : _V) {
            mean += v;
            meanOfLog += Math.log(v);
        }
        mean /= _V.length;
        logOfMean = Math.log(mean);
        meanOfLog /= _V.length;
        double shape = 0.5 / (logOfMean - meanOfLog);
        double scale = mean / shape;
        double logL_old, logL = _V.length * (-Gamma.logGamma(shape) - shape * Math.log(scale) - mean / scale + (shape - 1) * meanOfLog);
        do {
            logL_old = logL;
            shape = 1. / (1. / shape + (meanOfLog - logOfMean + Math.log(shape) - digamma(shape)) / (shape * shape * (1 / shape - trigamma(shape))));
            scale = mean / shape;
            logL = _V.length * (-Gamma.logGamma(shape) - shape * Math.log(scale) - mean / scale + (shape - 1) * meanOfLog);
        } while (1e-5 < Math.abs(logL - logL_old));//The threshold is set taking the accuracy of digamma and trigamma into account.
        return new DoublePair(logL, new double[]{shape, scale});
    }
}
