package fit.asympdist;

import fit.utils.DoublePair;
import java.util.Arrays;
import org.apache.commons.math3.distribution.LogNormalDistribution;

/**
 * Lognormal distribution.
 */
public class AsympDist_LogNormal implements AsympDist {

    /**
     * Constant frequently used in MLE calculation.
     */
    private static final double Q = 0.5 * Math.log(2. * Math.PI);

    @Override
    public int paramLen() {
        return 2;
    }

    /**
     * Returns MLE-estimated parameters and maximum log likelihood.
     *
     * @param _V Random variables.
     * @return logML, {mean, sd}.
     */
    @Override
    public DoublePair MLE(double[] _V) {
        double[] log = Arrays.stream(_V).map(v -> Math.log(v)).toArray();
        double meanLog = Arrays.stream(log).average().getAsDouble();
        double sdLog = Math.sqrt(Arrays.stream(log).map(v -> v - meanLog).map(v -> v * v).average().getAsDouble());
        double l = -_V.length * (meanLog + Math.log(sdLog) + Q + 0.5);
        return new DoublePair(l, new double[]{meanLog, sdLog});
    }

    /**
     * Inversed CDF.
     *
     * @param _p Cumulative probability value.
     * @param _Params {meanLog, sdLog}
     */
    @Override
    public double InvCDF(double _p, double... _Params) {
        return new LogNormalDistribution(_Params[0], _Params[1]).inverseCumulativeProbability(_p);
    }

    /**
     * Returns logarithmic PDF.
     *
     * @param _x Random variable.
     * @param _Params {mean, sd}
     * @return logPDF
     */
    @Override
    public double logPDF(double _x, double... _Params) {
        double logX = Math.log(_x);
        return -(logX + Math.log(_Params[1]) + Q + 0.5 * Math.pow((logX - _Params[0]) / _Params[1], 2.));
    }
}
