package fit.growth;

import org.apache.commons.math3.analysis.solvers.BrentSolver;
import org.apache.commons.math3.exception.NoBracketingException;
import org.apache.commons.math3.exception.TooManyEvaluationsException;

/**
 * Interface for growth models.
 */
public interface Growth {

    /**
     * Growth curve: V = X(_t; _Asymptote, _growthParams)
     *
     * @param _t Age
     * @param _Asymptote Asymptote
     * @param _growthParams Growth parameters except for the aymptote
     * @return Individual size
     */
    public double calcXAtT(double _t, double _Asymptote, double... _growthParams);

    /**
     * ln(dA/dX) = -ln(dX/dA), where A is asymptote and X is the growth function
     *
     * @param _t Age.
     * @param _Asymptote Asymptote
     * @param _growthParams Growth parameters
     * @return -ln(dX/dA)
     */
    public double log_dAdX( double _t, double _Asymptote, double... _growthParams);

    /**
     * Returns parameter length.
     * @return Number of parameters.
     */
    public int paramLen();

    /**
     * Calcuate the aymptote for _t, _x datum with given growth parameters.
     * @param _t Age.
     * @param _x Height.
     * @param _growthParams Growth parameters.
     * @return Asymptotic values.
     */
    public default double getAsymptote(double _t, double _x, double... _growthParams) {
        try {
            double res = new BrentSolver().solve(10000,
                    s -> _x - calcXAtT(_t, s, _growthParams),
                    0, 1e+4, _x);
            return res;
        } catch (NoBracketingException | TooManyEvaluationsException e) {
            return Double.NaN;
        }
    }

}
