package fit.utils;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.stream.IntStream;

/**
 * Class for handling files.
 */
public class FileHandler {

    public static String[] getFileAsStringArray(String _filename) throws IOException {
        File file = new File(_filename);
        byte[] Buf = new byte[(int) file.length()];
        try (DataInputStream dis = new DataInputStream(new FileInputStream(file))) {
            dis.readFully(Buf);
            dis.close();
        }
        return new String(ByteBuffer.wrap(Buf).array()).split("\r\n|[\n\r\u2028\u2029\u0085]");
    }

    /**
     * <pre>
     * Imports data from a file. The file should be formatted
     * Row 1: T, H (header)
     * Row 2: 50, 10 (datum) ...
     * </pre>
     *
     * @param _filename File path.
     * @return Set of age and size, i.e. {{age,size}, {age,size}, ...}
     * @throws java.io.IOException If the file could not read correctly.
     */
    public static double[][] getData(String _filename) throws IOException {
        String[] S = getFileAsStringArray(_filename);
        return Arrays.stream(S, 1, S.length).map(s -> s.split(","))
                .map(s -> new double[]{Double.parseDouble(s[0]), Double.parseDouble(s[1])})
                .toArray(double[][]::new);
    }

    /**
     * Generates summary of the fitting.
     *
     * @param _id ID of this result
     * @param _pso The executed PSO.
     * @return String Summary to be saved.
     */
    public static String getResult(int _id, PSO _pso) {
        double[] growthParams = _pso.getResult();// Height curve parameters
        double[] Asymptotes = _pso.fit.getAsymptotes(_pso.getResult());//Asymptotes of each datum
        double[] AsympDistParams = _pso.fit.asympDist.MLE(Asymptotes).params;//Distribution parameter of the asymptote
        return "ID=," + _id//Splitted with comma for csv file
                + ",Type=," + _pso.fit.getType()
                + ",iter=," + _pso.getNumIter()
                + ",-ln(LH)=," + _pso.getObjVal()
                + ",AIC=," + _pso.fit.getAIC(-_pso.getObjVal())
                + ",Hmax center=," + _pso.fit.asympDist.InvCDF(0.5, AsympDistParams)
                + ",AsympDistParams=," + myArraysToString(AsympDistParams)
                + ",CurveParams=," + myArraysToString(growthParams) + "\n";
    }

    /**
     * Get model parameters from Strings generated using getResult.
     *
     * @param _summary Strings generated with getResult.
     * @return {AsympDistParams, CurveParams}
     */
    public static double[][] getParams(String _summary) {
        String[] sp = _summary.split(",");
        int idx1 = -1, idx2 = -1;
        for (int i = sp.length - 1; -1 < i; i--) {
            if (sp[i].startsWith("AsympDistParams")) idx1 = i;
            else if (sp[i].startsWith("CurveParams")) idx2 = i;
            if (-1 < idx1 && -1 < idx2) break;
        }
        if (Math.min(idx1, idx2) < 0) return null;
        double[] AsympDistParams = IntStream.range(idx1 + 1, idx2).mapToDouble(i -> Double.parseDouble(sp[i])).toArray();
        double[] CurveParams = IntStream.range(idx2 + 1, sp.length).mapToDouble(i -> Double.parseDouble(sp[i])).toArray();
        return new double[][]{AsympDistParams, CurveParams};
    }

    /**
     * Returns String array without brackets.
     *
     * @param _ar Array to convert.
     * @return String of the array.
     */
    public static String myArraysToString(double[] _ar) {
        return Arrays.toString(_ar).replace("[", "").replace("]", "");
    }

    /**
     * Returns String array without brackets.
     *
     * @param _ar Array to convert.
     * @return String of the array.
     */
    public static String myArraysToString(int[] _ar) {
        return Arrays.toString(_ar).replace("[", "").replace("]", "");
    }
}
