package fit.utils;

import fit.Fit;
import java.util.Arrays;
import java.util.stream.IntStream;

/**
 * The particle swarm optimization of Fourie & Groenwold (2002), with parameters provided by Schutte & Groenwold (2005).
 */
public class PSO implements Runnable {

    //Fixed meta parameters of PSO.
    public static final int maxIteration = 1000;// Maximum iteration.
    public static final double stopDiff = 1e-8;//When the difference between the minimum and median particles in terms of the objective value is less than this value, the PSO is finished.
    public static final int nParticles = 56; //Number of particles.
    public static final double gamma = 0.4, c1 = 2, c2 = 2, alpha = 0.99, beta = 0.99;
    public static final int h = 10;

    private double objval;//Minimum value of objective function that has been obtained by any particles.
    private final double[] Solution;//The best position found by the swarm so far.
    private double w = 1.4; //Inertia weight; a controled meta parameter of PSO.
    private final double[] fQkh;//Values of the objective function at the past h.
    private int iter = 0;//Executed number of iterations.

    private final double[] MaxVelo; //Maximum velocities.
    private final double[][] BestPosEach;  //T best position found by each particle so far.
    private final double[][] CurrentPos;  //Current position of each particle.
    private final double[][] Velocities;  //Current velocity of each particle.
    private final double[] CurrentVal;  //Current value of objective function of each particle.
    private final double[] BestValues;  //Minimum value found by each particle so far.
    private final double[] BestPosAll; //Current best position of all particles at the iteration.
    public final Fit fit; //Objective function. Shallow copied.

    /**
     * Dimension of search space.
     */
    public final int nParams;

    /**
     * Constructor.
     *
     * @param _fit Fit class.
     */
    public PSO(Fit _fit) {
        fit = _fit;//Shallow copy
        nParams = fit.growth.paramLen();
        //Generates arrays.
        Solution = new double[nParams];
        MaxVelo = new double[nParams];
        BestPosEach = new double[nParticles][nParams];
        CurrentPos = new double[nParticles][nParams];
        Velocities = new double[nParticles][nParams];
        CurrentVal = new double[nParticles];
        BestValues = new double[nParticles];
        BestPosAll = new double[nParams];
        fQkh = new double[h];
    }

    /**
     * Updates the properties of the swarm and member particles.
     */
    private void updateSwarm(int _iter) {
        IntStream.range(0, nParticles).parallel().forEach(i -> {
            CurrentVal[i] = fit.f(CurrentPos[i]);
            if (CurrentVal[i] < BestValues[i]) {
                BestValues[i] = CurrentVal[i];
                System.arraycopy(CurrentPos[i], 0, BestPosEach[i], 0, CurrentPos[i].length);
            }
        });
        int minIndex = argmin(CurrentVal);
        //Shift the past values
        for (int i = h - 1; 0 < i; i--) fQkh[i] = fQkh[i - 1];
        fQkh[0] = CurrentVal[minIndex];
        System.arraycopy(CurrentPos[minIndex], 0, BestPosAll, 0, nParams);
        if (CurrentVal[minIndex] < objval) {
            objval = CurrentVal[minIndex];
            System.arraycopy(BestPosAll, 0, Solution, 0, nParams);
        }
    }

    /**
     * Initilizes the state of the PSO.
     *
     * @param _r Random number generator.
     */
    private void initialize(Rand _r) {
        w = 1.4;
        objval = Double.POSITIVE_INFINITY;
        Arrays.fill(Solution, Double.NaN);
        Arrays.fill(BestValues, Double.POSITIVE_INFINITY);
        Arrays.fill(fQkh, Double.POSITIVE_INFINITY);
        double range;
        for (int i = 0; i < nParticles; i++) {
            for (int j = 0; j < nParams; j++) {
                range = fit.Max[j] - fit.Min[j];
                BestPosEach[i][j] = CurrentPos[i][j] = range * _r.runif() + fit.Min[j];
                MaxVelo[j] = gamma * Math.abs(range);
                Velocities[i][j] = (_r.runif() < 0.5 ? -1 : 1) * MaxVelo[j];
            }
        }
        updateSwarm(0);
    }

    /**
     * Runs the PSO process.
     *
     */
    @Override
    public void run() {
        Rand r = new Rand();
        initialize(r);
        //Arrays to keep random numbers of each particles.
        double[] R1 = new double[nParticles], R2 = new double[nParticles];
        iter = 1;
        for (; iter < maxIteration; iter++) {
            r.fill(R1);
            r.fill(R2);
            IntStream.range(0, nParticles).parallel().forEach(i -> {
                double tmp;
                for (int j = 0; j < CurrentPos[i].length; j++) {
                    //The standard velocity rule
                    tmp = w * Velocities[i][j]
                            + c1 * R1[i] * (BestPosEach[i][j] - CurrentPos[i][j])
                            + c2 * R2[i] * (BestPosAll[j] - CurrentPos[i][j]);
                    Velocities[i][j] = (MaxVelo[j] < Math.abs(tmp)) ? (Math.signum(tmp) * MaxVelo[j]) : tmp;//Restriction of maximum velocity
                    tmp = CurrentPos[i][j] + Velocities[i][j];
                    CurrentPos[i][j] = Math.min(fit.Max[j], Math.max(fit.Min[j], tmp));//Limit the range
                }
            });
            updateSwarm(iter);
            if ((nParticles / 2 < Arrays.stream(BestValues).filter(d -> d - objval < stopDiff).count())) break;
            if (fQkh[0] >= fQkh[h - 1]) {//Update the inertia weight and maximum velocity.
                w *= alpha;
                for (int i = 0; i < nParams; i++) MaxVelo[i] *= beta;
            }
        }
    }

    /**
     * Returns the index of minimum number.
     *
     * @param Ar The array to identify the index.
     * @return Index of the minimum value.
     */
    private int argmin(double[] Ar) {
        double v = Ar[0];
        int idx = 0;
        for (int i = 1; i < Ar.length; i++) {
            if (Ar[i] < v) {
                v = Ar[i];
                idx = i;
            }
        }
        return idx;
    }

    /**
     * Returns the minimun objective function.
     *
     * @return Minimum objective function found.
     */
    public double getObjVal() {
        return objval;
    }

    /**
     * Returns the optimal decision variables as an array.
     *
     * @return The otimal variable set.
     */
    public double[] getResult() {
        return Solution;
    }

    /**
     * Returns the number of iterations.
     *
     * @return Executed number of iterations.
     */
    public int getNumIter() {
        return iter;
    }
}
