package main;

import fit.Fit;
import fit.asympdist.AsympDist_GGD;
import fit.asympdist.AsympDist_LogNormal;
import fit.growth.Growth_Anamorphic;
import fit.growth.Growth_Polymorphic;
import fit.utils.FileHandler;
import fit.utils.PSO;
import fit.utils.Rand;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.stream.IntStream;

/**
 * Class to do emperical bootstrap for estimating the confidence interval. Use many core PC for bootstrapping. Please use a PC with many cores.
 * The file to be prepared should be a csv file constructed as follows:
 * <pre>
 *
 * T,H The first row should be teh header
 * 30,20.4
 * 14,13.3
 * 25,20.0
 * ...
 *
 * </pre> Do not include row that includes missing data.
 */
public class Bootstrap {

    /**
     * The number of bootstrap sample.
     */
    public static final int N_BOOTSTRAP = 1001;

    /**
     * Set absolute file path of prepared filename (or place the file in the project directory). Fitting results are saved as csv files with prefix "ParamsOf_".
     */
    static final String filename = "Katakura2005_hinoki.csv";//Switch by commentout
    //static final String filename = "Katakura2005_larch.csv";//Switch by commentout

    public static void main(String... _args) throws IOException {
        System.setProperty("java.util.concurrent.ForkJoinPool.common.parallelism", "28");
        //Fitting methods to be attempted
        Fit[] F = {
            //The anamorphic model. Lower and upper limits of k and m.
            new Fit(new Growth_Anamorphic(), new AsympDist_LogNormal(), new double[]{1e-12, 0.01}, new double[]{0.1, 5}),
            new Fit(new Growth_Anamorphic(), new AsympDist_GGD(), new double[]{1e-12, 0.01}, new double[]{0.1, 5}),
            //The polymorphic model. Lower and upper limits of k, m and l.
            new Fit(new Growth_Polymorphic(), new AsympDist_LogNormal(), new double[]{1e-12, 0.01, -10}, new double[]{0.1, 5, 10}),
            new Fit(new Growth_Polymorphic(), new AsympDist_GGD(), new double[]{1e-12, 0.01, -10}, new double[]{0.1, 5, 10})
        };
        StringBuilder[] Sb = new StringBuilder[F.length];
        for (int i = 0; i < Sb.length; i++) Sb[i] = new StringBuilder(1024 * 1024);//For file save
        Rand r = new Rand();
        //Store original data for boot strap
        double[][] Original = FileHandler.getData(filename);
        double[][] BootStrapSample = new double[Original.length][2];
        //The first sample is original data
        for (int i = 0; i < Original.length; i++) System.arraycopy(Original[i], 0, BootStrapSample[i], 0, Original[i].length);

        for (int ID = 0; ID <= N_BOOTSTRAP; ID++) {
            //Sampling for boot strapping
            for (double[] sample : BootStrapSample) System.arraycopy(Original[Math.min((int) (r.runif() * Original.length), Original.length - 1)], 0, sample, 0, 2);
            for (int i = 0; i < F.length; i++) {
                F[i].setData(BootStrapSample);
                PSO pso = new PSO(F[i]);
                pso.run();
                String str = FileHandler.getResult(ID, pso);
                Sb[i].append(str);//First sample will not be recorded in a file (Only printed out)
                System.out.print(str);// Show result in the standard output
            }
            System.out.println();// Insert space for readability
        }
        for (int i = 0; i < F.length; i++) {
            String paramFileName = "ParamsOf_" + F[i].getType() + "_" + filename;
            Sb[i].delete(Sb[i].length() - 1, Sb[i].length());//Eliminate last /n
            try (FileWriter fw = new FileWriter(new File(paramFileName))) {
                fw.write(Sb[i].toString());
            }
            saveConfidenceIntervalCurves(F[i], paramFileName);
        }
    }

    public static void saveConfidenceIntervalCurves(Fit _f, String _filename) throws IOException {
        int[] Ages = IntStream.rangeClosed(10, 110).toArray();//Ages to be calculated
        double[] CurvePercentiles = {0.025, 0.5, 0.975};// Curves to be estimated
        double[] ConfidenceLevels = {0.025, 0.975};// Confidence levels to be calculated

        String[] Data = FileHandler.getFileAsStringArray(_filename);
        String[] Res = new String[Ages.length];
        String header = "Ages";
        for (double c1 : CurvePercentiles) for (double c2 : ConfidenceLevels) header += "," + "CurvePer:" + c1 + " ConfLev:" + c2;
        for (int i = 0; i < Ages.length; i++) Res[i] = "" + Ages[i];

        for (double c1 : CurvePercentiles) {
            int[] Indexes = Arrays.stream(ConfidenceLevels).mapToInt(d -> (int) (d * Data.length)).toArray();
            double[][] Estimates = new double[Data.length][];
            for (int i = 0; i < Data.length; i++) {
                double[][] Params = FileHandler.getParams(Data[i]);
                Estimates[i] = _f.getEstimates(Ages, Params[0], Params[1], c1);
            }
            double[][] Transposed = new double[Ages.length][Data.length];
            for (int i = 0; i < Data.length; i++) {
                for (int j = 0; j < Ages.length; j++) Transposed[j][i] = Estimates[i][j];
            }
            for (int i = 0; i < Ages.length; i++) {
                Arrays.sort(Transposed[i]);
                for (int j = 0; j < Indexes.length; j++) Res[i] += "," + Transposed[i][Indexes[j]];
            }
        }
        try (FileWriter fw = new FileWriter(new File("IntervalsOf_" + _filename))) {
            fw.write(header);
            for (String s : Res) {
                fw.write("\n");
                fw.write(s);
            }
        }
    }

}
