package main;

import fit.Fit;
import fit.asympdist.AsympDist_GGD;
import fit.asympdist.AsympDist_LogNormal;
import fit.growth.Growth_Anamorphic;
import fit.growth.Growth_Polymorphic;
import fit.utils.FileHandler;
import fit.utils.PSO;
import fit.utils.Rand;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import static main.Bootstrap.N_BOOTSTRAP;

/**
 * Reproduction test of "true" models.  Please use a PC with many cores for execution.
 */
public class Reproduction {

    //Log file name.
    static final String filename = "Reproduction_log.txt";

    //Polymorphic model with log-normal distribution for asymptotes
    private static final Fit PL = new Fit(new Growth_Polymorphic(), new AsympDist_LogNormal(), new double[]{1e-12, 0.01, -10}, new double[]{0.1, 5, 10});
    //Polymorphic model with generalized gamma distribution for asymptotes
    private static final Fit PG = new Fit(new Growth_Polymorphic(), new AsympDist_GGD(), new double[]{1e-12, 0.01, -10}, new double[]{0.1, 5, 10});
    //Polymorphic model with log-normal distribution for asymptotes
    private static final Fit AG = new Fit(new Growth_Anamorphic(), new AsympDist_GGD(), new double[]{1e-12, 0.01}, new double[]{0.1, 5});

    //"True" parameters {Asymptote distribution, Growth curve set}
    private static final double[][] PL_Cypress = {{3.25219309923104, 0.0307683453346914}, {0.0189825091976404, 1.16487551631949, 8.28628662094911}};
    private static final double[][] PG_Cypress = {{30.9789189213613, 3.24690866169441, 8.69912814392709}, {0.027820602531555, 1.21773689728085, 3.85037139097956}};
    private static final double[][] AG_Cypress = {{4.72973686360326, 2.47169849990733, 11.9094958039601}, {0.0413156510542587, 1.49678924473041}};

    private static final double[][] PL_Larch = {{3.26412279483286, 0.123391727594032}, {0.0448186525139554, 1.32238698872101, 0.754306595558983}};
    private static final double[][] PG_Larch = {{2.03085610654798, 5.48326467518244, 24.1153097339872}, {0.0452208079985849, 1.29944394414726, 0.269180583394303}};
    private static final double[][] AG_Larch = {{1.70167060894772, 5.48098333950487, 24.9102406162809}, {0.0467198786027658, 1.29168259934587}};

    //A data class for simple design of each examination condition
    static class ExamData {

        final String name; //Examination identifier
        final Fit True; //True model
        final double[] TrueCurveParam;//Parameters of growth curve set of the true model
        final double[] TrueDistParam;//Parameters of asymptote distribution of the true model
        final Fit Estimator;//Model assumed for estimation

        //Constructor
        ExamData(String _name, Fit _True, double[][] _TrueParam, Fit _Estimator) {
            name = _name;
            True = _True;
            TrueDistParam = _TrueParam[0];
            TrueCurveParam = _TrueParam[1];
            Estimator = _Estimator;
        }

    }

    //All examinations to be executed
    static ExamData[] ED = new ExamData[]{
        //Match patterns
        new ExamData("PL_Cypress_PL", PL, PL_Cypress, PL), //Estimate PL model of cypress data with PL model
        new ExamData("PG_Cypress_PG", PG, PG_Cypress, PG),
        new ExamData("PL_Larch_PL", PL, PL_Larch, PL),
        new ExamData("PG_Larch_PG", PG, PG_Larch, PG),
        new ExamData("AG_Cypress_AG", AG, AG_Cypress, AG),
        new ExamData("AG_Larch_AG", AG, AG_Larch, AG),

        //Miss match patterns
        new ExamData("PL_Cypress_PG", PL, PL_Cypress, PG), //Estimate PL model of cypress data with PG model
        new ExamData("PG_Cypress_PL", PG, PG_Cypress, PL),
        new ExamData("PL_Larch_PG", PL, PL_Larch, PG),
        new ExamData("PG_Larch_PL", PG, PG_Larch, PL),
        new ExamData("AG_Cypress_PG", AG, AG_Cypress, PG),
        new ExamData("PG_Cypress_AG", PG, PG_Cypress, AG),
        new ExamData("AG_Larch_PG", AG, AG_Larch, PG),
        new ExamData("PG_Larch_AG", PG, PG_Larch, AG),
    };

    //Number of samples per estimation
    private static final int[] N_SAMPLES = {100, 200};

    //Percentile asymptotes saved for evaluation
    private static final double[] CurvePercentiles = {0.025, 0.5, 0.975};

    //Confidence intervals of each curve saved for evaluation
    private static final double[] EstimationConfIntervals = {0.025, 0.5, 0.975};

    //Random value generator
    private static Rand r = new Rand();

    public static void main(String[] _args) throws IOException {

        System.setProperty("java.util.concurrent.ForkJoinPool.common.parallelism", "28");
        for (int n : N_SAMPLES) {
            for (ExamData E : ED) {
                StringBuilder sb = new StringBuilder(1024 * 1024);//Strings saved as file
                double[][][] Estimates = new double[N_BOOTSTRAP][CurvePercentiles.length][110];//Estimated values. Each index of the last dimension indicates age.
                for (int ID = 0; ID < N_BOOTSTRAP; ID++) {
                    double[][] Sample = new double[n][2];
                    generateVirtualData(r, Sample, E);
                    E.Estimator.setData(Sample);
                    PSO pso = new PSO(E.Estimator);
                    pso.run();
                    double[] CurveParams = pso.getResult();
                    fillEstimates(Estimates[ID], CurveParams, E.Estimator);
                    String str = FileHandler.getResult(ID, pso);
                    sb.append(str);
                    System.out.print(str);// Show result in the standard output
                    System.out.println();// Insert space for readability
                }
                sb.append(orderEstimates(Estimates, E));
                try (FileWriter fw = new FileWriter(new File("Z:\\" + E.name + n + ".csv"))) {
                    fw.write(sb.toString());
                }
            }
        }
    }

    //Summarize the estimation test by generating predictions for each percentile curves
    static StringBuilder orderEstimates(double[][][] _Estimates, ExamData _E) {
        StringBuilder sb = new StringBuilder(1024 * 1024);
        sb.append("\n\nAge");
        for (double c : CurvePercentiles) {
            sb.append(",Curve P:" + c + " True");
            for (double e : EstimationConfIntervals) sb.append(",Curve P:" + c + " Est conf:" + e);
            sb.append(",");
        }
        double[] Values = new double[N_BOOTSTRAP];
        for (int t = 0; t < 110; t++) {
            sb.append("\n" + (t + 1));//Age is index + 1
            for (int p = 0; p < CurvePercentiles.length; p++) {
                sb.append("," + _E.True.getEstimates(new int[]{t + 1}, _E.TrueDistParam, _E.TrueCurveParam, CurvePercentiles[p])[0]);//Growth curves of "true" model. Age is index + 1
                for (int k = 0; k < N_BOOTSTRAP; k++) Values[k] = _Estimates[k][p][t];
                Arrays.sort(Values);
                for (double ep : EstimationConfIntervals) sb.append("," + Values[(int) (N_BOOTSTRAP * ep)]);
                sb.append(",");
            }
        }
        return sb;
    }

    /**
     * Record estimated size for each percentile and age.
     *
     * @param _Estimates Array for recording estimations. First dimension indicates the percentile. Index values of second dimention (+ 1) directly indicates
     * the age.
     * @param _CurveParams Estimated parameters of growth curve set.
     * @param _Estimator Estimation model.
     */
    static void fillEstimates(double[][] _Estimates, double[] _CurveParams, Fit _Estimator) {
        double[] Asymptotes = _Estimator.getAsymptotes(_CurveParams);           //Asymptotes of each datum
        double[] AsympDistParams = _Estimator.asympDist.MLE(Asymptotes).params; //Distribution parameter of the asymptote
        for (int i = 0; i < _Estimates.length; i++) {
            double asymptote = _Estimator.asympDist.InvCDF(CurvePercentiles[i], AsympDistParams);
            for (int j = 0; j < _Estimates[i].length; j++) {
                _Estimates[i][j] = _Estimator.growth.calcXAtT(j + 1, asymptote, _CurveParams);
            }
        }
    }

    /**
     * Generate virtual data from "true" model.
     *
     * @param _r Sourse of random values.
     * @param _Sample Generated data will be placed in this array.
     * @param _E "True" properties in this ExamData are used.
     */
    private static void generateVirtualData(Rand _r, double[][] _Sample, ExamData _E) {
        for (double[] b : _Sample) {
            b[0] = r.runif() * 100 + 10; //Generate random ages in [10, 110]
            b[1] = _E.True.growth.calcXAtT(b[0], _E.True.asympDist.InvCDF(_r.runif(), _E.TrueDistParam), _E.TrueCurveParam);//Generate a size randomly
        }
    }

}
